%
% constraints.pl (Automatische Stundenplangenerierung)
% 
% Enthlt die Prdikate zum Erstellen der allgemeinen Constraints eines Stundenplanes
% (Zeit-, Raum- und Blockconstraints)
%

:- use_module(library(fd)).

/******************************************************************************************/
/*  Constraints bzgl. der Rume ...                                                       */
/******************************************************************************************/

/******************************************************************************************/
%
% roomConstraints/2(+Vorlesungen,+Vars)
%
% Wendet auf alle Vorlesungen der +Vorlesungen Liste die roomConstraints/2 an.
%
% +Vorlesungen Liste von vorlesung/4-Strukturen die die Vorlesungen enthlt, deren 
% Raum-Constraintvariable eingeschrnkt werden sollen.
% +Vars die dtr/4-Constraintvariablenliste, die die Problemvariable enthlt.
/******************************************************************************************/
roomConstraints([],_).                                        
roomConstraints([Course|Courses],Vars) :-
        roomConstraints(Course ,Vars),!,
        roomConstraints(Courses,Vars).

/******************************************************************************************/
%
% roomConstraints/2(+Vorlesung,+Vars)
%
% Wendet auf alle Raum-Constraintvariable die zur bergebenen Vorlesung gehren (befinden sich
% in Vars), ein Constraint ans, so da der Wertebereich dieser Raumvariablen nur noch Rume
% umfasst, die die Resourcenanforderungen der Vorlesung erfllen.
%
% +Vorlesung vorlesung/4-Struktur die die Vorlesung enthlt, deren Raum-Constraintvariable 
%            eingeschrnkt werden sollen.
% +Vars die dtr/4-Constraintvariablenliste, die die Problemvariable enthlt.
/******************************************************************************************/
roomConstraints(vorlesung(Name,_,Resources,_),Vars) :-
	checkRoomResources(Resources,Rooms),
	!,roomsAvailable(Rooms,Name),
	roomNumbers(RoomNumbers),
	names2Numbers(Rooms,RoomNumbers,Numbers),
	blockNumbers(Blocks),
	findall(BID,member(block(BID,Name,_),Blocks),BList),
	applyRoomConstraints(BList,Vars,Numbers).
	
%
% roomsAvailable/2(+RaumListe,+Vorlesungsname)
%
% Success wenn die Liste der Rume +Raumliste die fr die Vorlesung mit dem Namen +Vorlesungsname
% in Frage kommen nicht leer ist. Sonst Failure unter Ausgabe einer Fehlermeldung.
%
roomsAvailable([_|_],_) :- !.
roomsAvailable(_,Name) :- 
    message(['Es konnte kein passender Raum fr Kurs : ',Name,' gefunden werden']),!,fail.

	
%
% applyRoomConstraints/3(+BIDList,+Vars,+RoomNumbers)
%
% Schrnkt den Wertebereich aller Raum-Constraintvariable aus +Vars (dtr/4-Liste) die zu
% den BlockID's aus der Liste +BIDList gehren auf die Menge der Rume die durch die Liste
% +RoomNumbers definiert wird, ein.
%
% +BIDList Liste von BID's
% +Vars    dtr/4-Constraintvariablenliste 
% +RoomNumbers Liste mit ID's fr Rume
%
applyRoomConstraints([],_,_).
applyRoomConstraints([BID|T],Vars,Numbers):-
    getRoomsForBID(BID,RoomList,Vars),
    applyRoomConstraints_2(RoomList,Numbers),
    applyRoomConstraints(T,Vars,Numbers).
    
%
% Hilfsprdikat fr applyRoomConstraints/3
%
applyRoomConstraints_2([],_).
applyRoomConstraints_2([Room|T],Numbers):-
    Room::Numbers,
    applyRoomConstraints_2(T,Numbers).

        
%
% getRoomForBID/3(+BID,?RoomList,+DTRList)
%
% Die Liste RoomList enthlt die Raum-Constraintvariablen aus der dtr/4-Constraintvariablenliste 
% die zur BlockID BID gehren
%
getRoomsForBID(_,[],[]).
getRoomsForBID(BID,RoomList,[dtr(_,_,R,BID)|T]) :-
    !,getRoomsForBID(BID,RoomList0,T),
    append([R],RoomList0,RoomList).
getRoomsForBID(BID,RoomList,[_|T]) :-
    getRoomsForBID(BID,RoomList,T).


%Liefert die Liste "Rooms" aller Rume die die geforderten Resourcen "Resources" haben 
checkRoomResources(Resources,Rooms) :-
	roomsHaveResources(Resources,Rooms).

/******************************************************************************************/
/*  Constraints bzgl. der Zeit                                                            */
/******************************************************************************************/
/******************************************************************************************/
%
% timeConstraints/1(+Vars)
%
% Erstellt lineare Constraints, so das die Blcke der Vorlesungspaare aus der nst/2-Liste
% commonGroups/1 nicht gleichzeitig stattfinden knnen. Die nst/2 Liste wurde unter dem 
% Prdikat commonGroups/1 von der Proc. notTheSameTime/2 in die Datenbasis geschrieben.
%
% +Vars dtr/4-Constraintvariablenliste des Stundenplanes ber deren Variablen die Constraints
%       aufgebaut werden
%
/******************************************************************************************/
timeConstraints(Vars):-
    commonGroups(Common),                      %dyn. Prdikat aus Datenbasis
    mkCommonSlotsList(Common,Vars,Slots),
    traverseCommonSlots(Slots).

%
% traverseCommonSlots/1(+NstDTRList)
%
% Baut ber allen dtr/4 Constraintvariablenstrukturpaaren der nst/2 Liste +NstDTRList Constraints
% auf, so da die entsprechenden Vorlesungen nicht parallel stattfinden knnen.
%    
traverseCommonSlots([]).
traverseCommonSlots([nst(dtr(Day1,Time1,_,_),dtr(Day2,Time2,_,_))|T]):-
     noOfTimes(NoOfTimes),
     traceError([' Fehler bei Constraint bzgl nichtdisjunkter Gruppen (traverseCommonSlots)']),
     NoOfTimes * Day1 + Time1 #\= NoOfTimes * Day2 + Time2,
     traverseCommonSlots(T).

%
% mkCommonSlotsList/3(+NSTList,+DTRList,+Slots)
%
% Wird von timeConstraints/1 benutzt um fr jedes Vorlesungspaar aus der nst/2-Liste die
% Liste der zugehrigen dtr/4-Constraintvariablenstrukturen zu erstellen, ber denen dann
% Constraints so gesetzt werden knnen, da die Vorlesungen nicht parallel stattfinden knnen
% 
mkCommonSlotsList([],_,[]).
mkCommonSlotsList([nst(A,B)|T],Vars,Slots) :-
    mkCommonVars(A,B,Vars,CommonVars), /*Liefert Liste aller dtr/4 - Variablenstrukturpaare, die nicht  */
                                       /*gleichzeitig stattfinden drfen. */
    mkCommonSlotsList(T,Vars,Rest),    /*Rekursion*/
    append(CommonVars,Rest,Slots).     /*Akk. des Ergebn. */
    
%  
% Liefert fr die beiden Vorlesungen A und B die nicht parallel stattfinden sollen die
% nst/2 Liste CommonVars aller dtr/4 Constraintvariablenstrukturenpaare die nich parallel stattfinden
% drfen. Vars ist die Liste aller dtr/4-Constraintvariablenstrukturen des Stundenplanes
%
mkCommonVars(A,B,Vars,CommonVars) :-
    blockNumbers(Blocks),
    findall(BID_A,member(block(BID_A,A,_),Blocks),BIDList_A),
    findall(BID_B,member(block(BID_B,B,_),Blocks),BIDList_B),
    getDTRsForBID_s(BIDList_A,Vars,[],DTRList_A),
    getDTRsForBID_s(BIDList_B,Vars,DTRList_A,DTRList),
    mkCommonVarPairs(DTRList,[],CommonVars).


%
% Liefert alle dtr/4Constraintvariablenstrukturpaare zu allen BlockIDs der BID-Liste
%
getDTRsForBID_s([],_,AccRes,AccRes).
getDTRsForBID_s([BIDH|T],Vars,AccResult,Result) :-
    getDTRsForBID(BIDH,Vars,AccResult,Result0),
    getDTRsForBID_s(T,Vars,Result0,Result).
    
%
% Liefert alle dtr/4-Constraintvaribalenstrukturpaare zur bergebenen BlockId BID
%
getDTRsForBID(_,[],AccRes,AccRes).
getDTRsForBID(BID,[dtr(D,T,R,BID)|VT],AccRes,Result) :-
    !,append([dtr(D,T,R,BID)],AccRes,Res),
    getDTRsForBID(BID,VT,Res,Result).
getDTRsForBID(BID,[_|VT],AccRes,Result):-
    getDTRsForBID(BID,VT,AccRes,Result).
    

%
% Liefert die Elemente der dtr/4-Constraintvariablenstrukturliste die Liste aller 
% ungeordneten Paare der dtr/4-Liste
%
mkCommonVarPairs([],Pairs,Pairs).
mkCommonVarPairs([dtr(W,X,Y,Z)|T],Pairs,Result) :-
    traverseTail(dtr(W,X,Y,Z),T,Pairs,TempPairs),
    mkCommonVarPairs(T,TempPairs,Result).

%
% Hilfsprdikat fr mkCommonVarPairs/3
%    
traverseTail(_,[],Res,Res).
traverseTail(dtr(W,X,Y,Z),[dtr(D,Ti,R,B)|T],Pairs,Result) :-
    append([nst(dtr(W,X,Y,Z),dtr(D,Ti,R,B))],Pairs,TempPairs),
    traverseTail(dtr(W,X,Y,Z),T,TempPairs,Result).
    
/******************************************************************************************/
/* Constraints bezglich der Blockung von Veranstaltungen                                 */
/*   1) Blcke mssen zeitlich "in einem Stck" stattfinden                               */
/*   2) Falls Blcke so definiert sind das sie nicht am selben Tag liegen drfen wie      */
/*      ein anderer Block derselben Vorlesung (Clusterbildung) muessen die entsprechenden */
/*      Constraints gesetzt werden.                                                       */
/******************************************************************************************/
blockConstraints(Vars) :-
    setupBlocks(Vars,-1,_),
    deClusterBlocks(Vars,Vars,-1).
    
setupBlocks([],_,_). 
setupBlocks([dtr(D,Ti,R,BID)|T],LastBID,dtr(Day,LastTime,Room)) :-
    LastBID=BID,
    ( 
      D#=Day,         /* Ein Block kann sich nicht ber mehr als einen Tag erstrecken */
      Ti#=LastTime+1, /* Vorlesungseinheiten eines Blockes folgen direkt aufeinander */
      R#=Room         /* Whrend eines Blockes wird der Raum nicht gewechselt */
    ), 
    setupBlocks(T,BID,dtr(Day,Ti,Room)).        
setupBlocks([dtr(D,Ti,R,BID)|T],LastBID,dtr(_,_,_)) :-
    LastBID\=BID,
    setupBlocks(T,BID,dtr(D,Ti,R)).
    
deClusterBlocks(_,[],_).
deClusterBlocks(Vars,[dtr(_,_,_,BID)|T],_):-
    mayCluster(BID),                                 /* Plausibilittscheck */
    deClusterBlocks(Vars,T,BID).
deClusterBlocks(Vars,[dtr(D,Ti,R,BID)|T],LastBID):-
    LastBID\=BID,
    not(mayCluster(BID)),
    getRestOfBlock(BID,T,Result,Rest),
    append([dtr(D,Ti,R,BID)],Result,Slots),
    deClusterBlock(Slots,Vars),
    deClusterBlocks(Vars,Rest,BID).
    
deClusterBlock(Slots,Vars):-
    Slots=[dtr(_,_,_,BID)|_],
    deClusterBlock(Slots,BID,Vars).

deClusterBlock(_,_,[]).				    /* BOUNDARY */
deClusterBlock(Slots,BID,[dtr(D,_,_,ActBID)|T]) :-  /* BID\=ActBID aber identische Vorlesung */
    BID\=ActBID,
    nameForBID(BID,Name),
    nameForBID(ActBID,Name),!,
    applyDeClusterConstraint(Slots,D),
    deClusterBlock(Slots,BID,T).
deClusterBlock(Slots,BID,[dtr(_,_,_,BID)|T]) :-     /* BID=BID */
    deClusterBlock(Slots,BID,T).
deClusterBlock(Slots,BID,[dtr(_,_,_,ActBID)|T]) :-  /* BID\=ActBID verschiedene Vorlesungen */
    BID\=ActBID,
    deClusterBlock(Slots,BID,T).
    
applyDeClusterConstraint([],_).
applyDeClusterConstraint([dtr(Day,_,_,_)|T],D) :-
    D#\=Day,
    applyDeClusterConstraint(T,D).
    
mayCluster(BID) :-
    blockNumbers(Blocks),
    member(block(BID,_,Structure),Blocks),
    (
      Structure=single;
      Structure=block2;
      Structure=block3
    ).
    
getRestOfBlock(_,[],[],[]).
getRestOfBlock(BID,[dtr(D,Ti,R,BlockID)|T],Result,Rest) :-
    BID\=BlockID,
    append([dtr(D,Ti,R,BlockID)],T,Rest),
    Result=[].
getRestOfBlock(BID,[dtr(D,Ti,R,BID)|T],Result,Rest) :-
    getRestOfBlock(BID,T,Result0,Rest),
    append([dtr(D,Ti,R,BID)],Result0,Result).
   
nameForBID(BID,Name) :-
    blockNumbers(Blocks),
    member(block(BID,Name,_),Blocks).
















