%                                                                            
% specificConstraints.pl (Automatische Stundenplangenerierung)
% 
% Prdikate zur Verarbeitung der spezifischen Constraints aus der Datenbasis.
% 


/*************************************************************************************/
%
% applySpecificConstraints/1(+Vars)
%
% Wendet alle spezifischen Constraints aus der Datenbasis auf die Constraintvariablen an
%
% +Vars Die Liste der Constraintvariablenstrukturen dtr/4 des Problems
%
/*************************************************************************************/
applySpecificConstraints(Vars) :-
    processLecturerConstraints(Vars),
    !,processLectureConstraints(Vars).

    
/****************************************/
/* Lecturer Constraints                 */
/****************************************/
processLecturerConstraints(Vars) :-
    findall(D,dozent(D),Dozenten),
    !,processAllLecturers(Dozenten,Vars).

processAllLecturers([],_).
processAllLecturers([Dozent|T],Vars) :-
    getAllDTRsForLecturer(Dozent,Vars,DTRList),

    /* DozentNichtAm */
    findall(Tag,dozentNichtAm(Dozent,Tag),Tage), ensureIsList(Tage,FlatTage),
    findall(TagID,(member(Tg,FlatTage),dayNumber(Tg,TagID)),TagIDList),
    !,processDozentNichtAm(DTRList,TagIDList),
	    
    /* DozentNichtUm */
    findall(Zeit,dozentNichtUm(Dozent,Zeit),Zeiten), ensureIsList(Zeiten,FlatZeiten),
    findall(ZeitID,(member(Zt,FlatZeiten),timeNumber(Zt,ZeitID)),ZeitIDList),
    !,processDozentNichtUm(DTRList,ZeitIDList),
    
    /* DozentNichtAmUm */
    findall(nichtAmUm(Am,Um),dozentNichtAmUm(Dozent,Am,Um),NichtAmUmListe),
    !,processDozentNichtAmUm(DTRList,NichtAmUmListe),
    
    !,processAllLecturers(T,Vars).
    
processDozentNichtAm([],_).
processDozentNichtAm([dtr(D,_,_,_)|T],TagIDList) :-
    excludeDays(D,TagIDList),
    processDozentNichtAm(T,TagIDList).
    
excludeDays(_,[]).
excludeDays(Day,[TagH|T]) :-
    traceError('ExcludeDays'),
    Day#\=TagH,
    excludeDays(Day,T).

processDozentNichtUm([],_).
processDozentNichtUm([dtr(_,Time,_,_)|T],ZeitIDList) :-
    excludeTimes(Time,ZeitIDList),
    processDozentNichtUm(T,ZeitIDList).

excludeTimes(_,[]).
excludeTimes(Time,[TimeH|T]) :-
    traceError('ExcludeTimes'),
    Time#\=TimeH,
    excludeTimes(Time,T).

processDozentNichtAmUm([],_).
processDozentNichtAmUm([dtr(D,Ti,_,_)|T],NichtAmUm) :-
    excludeTimesAndDays(D,Ti,NichtAmUm),
    processDozentNichtAmUm(T,NichtAmUm).
     
excludeTimesAndDays(_,_,[]).
excludeTimesAndDays(Day,Time,[nichtAmUm(Am,Um)|T]) :-
    Um=[_|_],!,
    excludeTimesAndDays(Day,Time,Am,Um),
    excludeTimesAndDays(Day,Time,T).
excludeTimesAndDays(Day,Time,[nichtAmUm(Am,Um)|T]) :-
    excludeTimesAndDays(Day,Time,[nichtAmUm(Am,[Um])|T]).
    
excludeTimesAndDays(_,_,_,[]).
excludeTimesAndDays(Day,Time,Am,[Um|T]) :-
    traceError('ExcludeTimesAndDays'),
    ( Day#=Am ) #=> ( Time #\= Um ),
    ( Time#=Um) #=> ( Day #\= Am ),
    excludeTimesAndDays(Day,Time,Am,T).

getAllDTRsForLecturer(_,[],[]).
getAllDTRsForLecturer(Dozent,[dtr(D,T,R,BID)|T],DTRList) :-
    groupTakesPart(Dozent,BID),!,
    getAllDTRsForLecturer(Dozent,T,DTRList0),
    append([dtr(D,T,R,BID)],DTRList0,DTRList).
getAllDTRsForLecturer(Dozent,[_|T],DTRList) :-
    getAllDTRsForLecturer(Dozent,T,DTRList).
    
groupTakesPart(Group,BID) :-
    blockNumbers(Blocks),
    preparedCourses(Courses),
    member(block(BID,VName,_),Blocks),
    member(vorlesung(VName,Groups,_,_),Courses),
    member(Group,Groups).

/*************************************************************/
/* Lecture Constraints                                       */
/*    vorlesungAm  ----!                                     */
/*    vorlesungUm      +-- Nur fr Vorlesungen die aus       */
/*                     !   nicht mehr als einem unabh. Block */
/*                     !   bestehen !!                       */
/*    vorlesungAmUm ---!                                     */
/*    vorlesungNichtAm                                       */
/*    vorlesungNichtUm                                       */
/*    vorlesungNichtAmUm                                     */
/*************************************************************/
processLectureConstraints(Vars) :-
    preparedCourses(Courses),
    !,processAllCourses(Courses,Vars),
      
    blockNumbers(BList),
    findall(nichtZH(VNameA,VNameB),vorlesungenNichtZusammenhaengend(VNameA,VNameB),NichtZHListe),
    findall(nichtZH(BIDA,BIDB),
            (
	      member(nichtZH(A,B),NichtZHListe),
	      member(block(BIDA,A,_),BList),
	      member(block(BIDB,B,_),BList)
	    ),
	    NichtZHBIDListe
           ),
    !,processNichtZH(NichtZHBIDListe,Vars),
    
    findall(NZHList,vorlesungenNichtZusammenhaengend(NZHList),NZHLists),
    findall(nichtZH(BID_A,BID_B),
            ( 
	      member(NZH_Liste,NZHLists),
	      getNZHPairs(NZH_Liste,Pairs),
	      member(nichtZH(A1,B1),Pairs),
	      member(block(BID_A,A1,_),BList),
	      member(block(BID_B,B1,_),BList)
	    ),
	    NichtZHBID_Liste
           ),
    !,processNichtZH(NichtZHBID_Liste,Vars),
    
    /* VorlesungenHintereinander */
    findall(consec(A2,B2),vorlesungenHintereinander(A2,B2),ConsecList),
    !,processConsecutives(ConsecList,Vars),
        
    /* VorlesungenAmSelbenTag */
    findall(selberTag(A3,B3),vorlesungenAmSelbenTag(A3,B3),STagList),
    !,processSameDay(STagList,Vars),
        
    /* VorlesungenNichtAmSelbenTag */
    findall(nSelberTag(A4,B4),vorlesungenNichtAmSelbenTag(A4,B4),NSTagList),
    !,processNotSameDay(NSTagList,Vars),
    
    /* alleBloeckeAmSelbenTag*/
    findall(VorlName,alleBloeckeAmSelbenTag(VorlName),VorlNamen),
    !,processAllBlocksSameDay(VorlNamen,Vars).


proc_VorlesungNichtAm(VName,DTRList) :-
   findall(Tag,vorlesungNichtAm(VName,Tag),Tage), ensureIsList(Tage,FlatTage),
   findall(TagID,(member(TagName,FlatTage),dayNumber(TagName,TagID)),TagIDList),
   processVorlesungNichtAm(DTRList,TagIDList).
   
proc_VorlesungNichtUm(VName,DTRList) :-
    findall(Zeit,vorlesungNichtUm(VName,Zeit),Zeiten), ensureIsList(Zeiten,FlatZeiten),
    findall(ZeitID,(member(ZeitName,FlatZeiten),timeNumber(ZeitName,ZeitID)),ZeitIDList),
    processVorlesungNichtUm(DTRList,ZeitIDList).

proc_VorlesungNichtAmUm(VName,DTRList):-
    findall(nichtAmUm(Am,Um),vorlesungNichtAmUm(VName,Am,Um),NichtAmUmListe),
    processVorlesungNichtAmUm(DTRList,NichtAmUmListe).
    
processAllCourses([],_).
processAllCourses([vorlesung(VName,_,_,Blocks)|T],Vars) :-
   getAllDTRsForCourse(VName,Vars,DTRList),
   !,proc_VorlesungNichtAm(VName,DTRList),
   !,proc_VorlesungNichtUm(VName,DTRList),    
   !,proc_VorlesungNichtAmUm(VName,DTRList),
   !,processVorlesungXX(VName,Blocks,DTRList),
   !,processAllCourses(T,Vars).

     
processVorlesungNichtAm([],_).
processVorlesungNichtAm([dtr(Day,_,_,_)|DTRT],TagIDList) :-
    excludeDays(Day,TagIDList),
    processVorlesungNichtAm(DTRT,TagIDList).
    
processVorlesungNichtUm([],_).
processVorlesungNichtUm([dtr(_,Time,_,_)|DTRT],ZeitIDList) :-
    excludeTimes(Time,ZeitIDList),
    processVorlesungNichtUm(DTRT,ZeitIDList).    
    
processVorlesungNichtAmUm([],_).
processVorlesungNichtAmUm([dtr(Day,Time,_,_)|DTRT],NichtAmUmListe):-
    excludeTimesAndDays(Day,Time,NichtAmUmListe),    
    processVorlesungNichtAmUm(DTRT,NichtAmUmListe).
    
processVorlesungXX(VName,Blocks,DTRList) :-
    length(Blocks,1),
    Blocks=[H|_],
    (H=single(1);H=singleNoCluster(1);
     H=block2(1);H=block2NoCluster(1);
     H=block3(1);H=block3NoCluster(1)),

    /* VorlesungAm   */
    !,processVorlesungAm(VName,DTRList),

    /* VorlesungUm   */
    !,processVorlesungUm(VName,DTRList),
 
    /* VorlesungAmUm */
    !,processVorlesungAmUm(VName,DTRList).
    
processVorlesungXX(_,_,_).


processVorlesungAm(VName,DTRList) :-
    vorlesungAm(VName,Tag),!,
    dayNumber(Tag,TagID),
    setDay(DTRList,TagID).    
processVorlesungAm(_,_).

/* Geht davon aus, da das erste Element der DTR-Liste der Kopf des */
/* Blockes ist !! KEINE UMSORTIERUNG DER DTRs WIE SIE AUS DER      */
/* CONSTRAINTVARIABLENLISTE KOMMEN !!                               */
processVorlesungUm(VName,[dtr(_,T,_,_)|_]) :-
    vorlesungUm(VName,Um),!,
    timeNumber(Um,TimeID),  
    traceError(['VorlesungUm ',VName,' ',Um]),
    T#=TimeID.
processVorlesungUm(_,_).

processVorlesungAmUm(VName,DTRList):-
    vorlesungAmUm(VName,Am,Um),!,
    dayNumber(Am,DayID),
    timeNumber(Um,TimeID),
    traceError(['VorlesungAmUm ',VName,' ',Am,' ',Um]),
    setDay(DTRList,DayID),
    DTRList=[dtr(_,T,_,_)|_],
    T#=TimeID.    
processVorlesungAmUm(_,_).

setDay([],_).
setDay([dtr(D,_,_,_)|T],TagID):-
    traceError('SetDay'),
    D#=TagID,
    setDay(T,TagID).

getAllDTRsForCourse(_,[],[]).
getAllDTRsForCourse(VName,[dtr(D,T,R,BID)|VarsT],DTRList) :-
    blockNumbers(Blocks),
    member(block(BID,VName,_),Blocks),!,
    getAllDTRsForCourse(VName,VarsT,DTRList0),
    append([dtr(D,T,R,BID)],DTRList0,DTRList).
getAllDTRsForCourse(VName,[_|VarsT],DTRList) :-
    getAllDTRsForCourse(VName,VarsT,DTRList).


/*****************************************************************************/

processNichtZH([],_).
processNichtZH([nichtZH(BIDA,BIDB)|T],Vars):-
    firstOccurrence(dtr(D,_,R,BIDA),Vars,dtr(_,TaStart,_,_)),
    lastOccurrence(dtr(D,_,R,BIDA),Vars,dtr(_,TaStop,_,_)),
    firstOccurrence(dtr(Db,_,Rb,BIDB),Vars,dtr(_,TbStart,_,_)),
    lastOccurrence(dtr(Db,_,Rb,BIDB),Vars,dtr(_,TbStop,_,_)),
    traceError('ProcessNichtZH'),
    (D#=Db) #=> (TbStop+1 #\=TaStart), (TbStop+1 #= TaStart) #=> (D#\=Db),
    (D#=Db) #=> (TaStop+1 #\=TbStart), (TaStop+1 #= TbStart) #=> (D#\=Db),
    processNichtZH(T,Vars).
    
/*****************************************************************************/

processConsecutives([],_).
processConsecutives([consec(A,B)|T],Vars) :-
    processConsecutives(A,B,Vars),
    processConsecutives(T,Vars).

/* Fr die beiden Vorlesungen mit den Namen A und B werden Constraints gesetzt, so da */
/* die Vorlesung B zeitlich direkt auf die Vorlesung A folgt.                          */
processConsecutives(A,B,Vars) :-
    vorlesung(A,_,_,[Block]), singleBlock(Block),
    vorlesung(B,_,_,[Block2]),singleBlock(Block2),!,
    blockNumbers(Blocks),
    member(block(BIDA,A,_),Blocks), member(block(BIDB,B,_),Blocks),
    lastOccurrence(dtr(_,_,_,BIDA),Vars,dtr(Da,Ta,_,_)),
    firstOccurrence(dtr(_,_,_,BIDB),Vars,dtr(Db,Tb,_,_)),
    traceError('processConsecutives'),
    Da#=Db,
    (Ta+1)#=Tb.
processConsecutives(_,_,_).
    
%
% Success wenn die bergebene Blockstruktur fr einen einzelnen Block steht.
%
sinlgeBlock(single(1)).
singleBlock(singleNoCluster(1)).
singleBlock(block2(1)).
singleBlock(block2NoCluster(1)).
singleBlock(block3(1)).
singleBlock(block3NoCluster(1)).

/***********************************************************************************/

processSameDay([],_).
processSameDay([selberTag(A,B)|T],Vars) :-
    processSameDay(A,B,Vars),
    processSameDay(T,Vars).
    
processSameDay(A,B,Vars) :-
    vorlesung(A,_,_,[Block]), singleBlock(Block),
    vorlesung(B,_,_,[Block2]),singleBlock(Block2),!,
    blockNumbers(Blocks),
    member(block(BIDA,A,_),Blocks), member(block(BIDB,B,_),Blocks),    
    firstOccurrence(dtr(_,_,_,BIDA),Vars,dtr(Da,_,_,_)),
    firstOccurrence(dtr(_,_,_,BIDB),Vars,dtr(Db,_,_,_)),
    traceError(['processSameDay ',A,' ',B]),
    Da#=Db.
processSameDay(_,_,_).

processNotSameDay([],_).
processNotSameDay([nSelberTag(A,B)|T],Vars) :-
    processNotSameDay(A,B,Vars),
    processNotSameDay(T,Vars).
    
processNotSameDay(A,B,Vars) :-
    vorlesung(A,_,_,[Block]), singleBlock(Block),
    vorlesung(B,_,_,[Block2]),singleBlock(Block2),!,
    blockNumbers(Blocks),
    member(block(BIDA,A,_),Blocks), member(block(BIDB,B,_),Blocks),    
    firstOccurrence(dtr(_,_,_,BIDA),Vars,dtr(Da,_,_,_)),
    firstOccurrence(dtr(_,_,_,BIDB),Vars,dtr(Db,_,_,_)),
    traceError(['processNotSameDay ',A,' ',B]),
    Da#\=Db.
processNotSameDay(_,_,_).

/***************************************************************************************/
%
% processAllBlocksSameDay/2(+VNames,+Vars) 
%
% Setzt fr jede Vorlesung aus der Liste von Vorlesungsnamen +VNames Constraints, so da
% alle Blcke der jeweiligen Vorlesung an einem Tag stattfinden mssen. 
%
/***************************************************************************************/
processAllBlocksSameDay([],_).
processAllBlocksSameDay([VName|T],Vars):-
    allBlocksSameDay1(VName,Vars),
    processAllBlocksSameDay(T,Vars).
    
allBlocksSameDay1(VName,Vars) :-
    blockNumbers(Blocks),
    findall(Day,(member(block(BID,VName,_),Blocks),member(dtr(Day,_,_,BID),Vars)),Days),
    Days=[FirstDay|RestDays],
    sameDay(FirstDay,RestDays).
    
sameDay(_,[]).
sameDay(Day0,[Day1|T]):-
    sameDay(Day0,T),
    traceError('sameDay'),    
    Day0#=Day1.
     
/***************************************/
/* Utility Procs                       */
/***************************************/
%
% ensureIsList/2(+InList,-OutList)
% Macht aus einer evtl. geschachtelten Liste (Baum) +InList eine flache 
% Liste -OutList aller Elemente der Eingabeliste
ensureIsList([],[]).
ensureIsList([InH|InT],OutList) :-
    InH=[_|_],!,
    ensureIsList(InH,List),
    ensureIsList(InT,OutList0),
    append(List,OutList0,OutList).
ensureIsList([InH|InT],OutList) :-
    ensureIsList(InT,OutList0),
    append([InH],OutList0,OutList).
    
  
/* Matcht das Erste Element der Liste das mit X matcht mit Y */
firstOccurrence(_,[],_) :- fail.
firstOccurrence(X,[X|_],X).
firstOccurrence(X,[_|T],Y) :-
    firstOccurrence(X,T,Y).
/* Matcht das Letzte Element der Liste das mit X matcht mit Y */
lastOccurrence(X,List,Y) :-
    reverse(List,Reversed),
    firstOccurrence(X,Reversed,Y).    

/* Bildet die Liste aller ungeordneten Paare nichtZH/2 aller Elemente der bergebenen Liste */
getNZHPairs([],[]).
getNZHPairs([VName|T],Pairs) :-
    getNZHPairs(T,Pairs0),
    findall(nichtZH(Name,VName),member(Name,T),Paare),
    append(Paare,Pairs0,Pairs).

%
% Zur Ausgabe von Fehlermeldungen. Der 1. Versuch ein traceError/1 Goal zu erfllen ist
% immer Erfolgreich. Ein Backtracking zurck ber dieses Goal fhrt zur Ausgabe der
% Fehlermeldung die als Parameter bergeben wurde.
%
traceError(_).
traceError(Msg) :-
    message([' Folgendes Constraint schlug fehl -> ',Msg]),!,fail.

































